//go:build go1.20
// +build go1.20

// protobuf the way I am using them, require GO 1.20. I think. I could be wrong.
// The Go Protocol Buffers library embeds a sync.Mutex within the MessageState struct to prevent unintended shallow copies of message structs
// this optionally (but it is the default) inserts a mutex into the struct generated by protoc

// go:generate autogenpb --proto file.proto

package main

import (
	"errors"
	"fmt"
	"os"
	"path/filepath"
	"strings"

	"github.com/alexflint/go-arg"
	"github.com/go-cmd/cmd"
	"go.wit.com/lib/gui/shell"
	"go.wit.com/log"
)

// sent via -ldflags
var VERSION string
var BUILDTIME string

// var fsort *os.File // the sort.pb.go output file

func main() {
	pp := arg.MustParse(&argv)

	var pb *Files
	pb = new(Files)

	if argv.Identify != "" {
		if err := doIdentify(argv.Identify); err != nil {
			badExit(err)
		}
		okExit("")
	}

	// you need a proto file
	if argv.Proto == "" {
		log.Info("you must provide --proto <filename>")
		pp.WriteHelp(os.Stdout)
		os.Exit(-1)
	}

	if !shell.Exists(argv.Proto) {
		log.Info("protobuf", argv.Proto, "is missing")
		os.Exit(-1)
	}

	if !strings.HasSuffix(argv.Proto, ".proto") {
		log.Info("protobuf", argv.Proto, "must end in .proto")
		os.Exit(-1)
	}

	if argv.Regret {
		// this will override the manditory Uuid checks
		os.Setenv("PROTOBUF_REGRET", "true")
	}

	pf := new(File)
	pb.Files = append(pb.Files, pf)
	pf.Filename = argv.Proto
	pf.IterMap = make(map[string]string)

	pf.Filebase = strings.TrimSuffix(argv.Proto, ".proto")

	// parse sort & marshal options from the .proto file
	// this goes through the .proto files and looks
	// for `autogenpb: ` lines
	if err := pf.protoParse(); err != nil {
		log.Info("autogenpb parse error:", err)
		badExit(err)
	}

	if pf.Bases == nil {
		badExit(fmt.Errorf("Base was nil. 'message %s {` did not exist", pf.Filebase))
	}
	if pf.Base == nil {
		badExit(fmt.Errorf("Base was nil. 'message %s {` did not exist", pf.Filebase))
	}

	// if you have gotten here, at least the .proto buf file is OK
	if argv.DryRun {
		// show the protobuf of the protobuf. It's like Inception
		pf.printMsgTable()
		okExit("")
	}

	// todo, look for go.work files
	if argv.GoSrc == "" {
		homeDir, _ := os.UserHomeDir()
		argv.GoSrc = filepath.Join(homeDir, "go/src")
	}

	if argv.GoPath == "" {
		pwd, _ := os.Getwd()
		argv.GoPath = strings.TrimPrefix(pwd, argv.GoSrc)
		argv.GoPath = strings.Trim(argv.GoPath, "/")
	}
	log.Info(argv.GoSrc, argv.GoPath)

	pwd, _ := os.Getwd()
	log.Info("pwd = ", pwd)

	if !shell.Exists("go.sum") {
		shell.RunQuiet([]string{"go-mod-clean"})
		if !shell.Exists("go.sum") {
			shell.RunQuiet([]string{"go", "mod", "init"})
			shell.RunQuiet([]string{"go", "mod", "tidy"})
			shell.RunQuiet([]string{"go", "mod", "edit", "-go=1.18"}) // TODO: pass this as ENV. verify protobuf version needed
		}
	}

	var packageName string
	var result cmd.Status
	var cmd []string
	if argv.Package == "" {
		// TODO: switch to using forgepb (expose the functions/logic from forgepb directly
		// it could be a bad idea to use forge.Init() here as forge needs this to build
		// switch to forgepb
		os.Setenv("GO111MODULE", "off") // keeps go list working if go version is back versioned for compatability
		cmd = []string{"go", "list", "-f", "'{{.Name}}'"}
		result = shell.RunQuiet(cmd)
		os.Unsetenv("GO111MODULE")

		packageName = strings.Join(result.Stdout, "\n")
		packageName = strings.TrimSpace(packageName)
		packageName = strings.Trim(packageName, "'")
		// log.Info("packageName == ", packageName)
	} else {
		packageName = argv.Package
	}
	pf.Package = packageName

	// try to make foo.pb.go with protoc if it's not here
	// this is helpful because the protoc-gen-go lines
	// are also annoying to code by hand
	checkCmd("protoc")
	checkCmd("protoc-gen-go")

	pf.Pbfilename = pf.Filebase + ".pb.go"
	// try to create the foo.pb.go file using protoc if it is not there
	if !shell.Exists(pf.Pbfilename) {
		checkCmd("protoc")
		checkCmd("protoc-gen-go")

		if err := pb.protocBuild(pf); err != nil {
			badExit(err)
		}

	}

	// try to add the Mutex to the pb.go file
	if err := pb.addMutex(pf); err != nil {
		badExit(err)
	}

	// if foo.pb.go still doesn't exist, protoc failed
	if !shell.Exists(pf.Pbfilename) {
		log.Info("protoc build error.", pf.Pbfilename)
		badExit(errors.New("failed to be created with protoc and proto-gen-go"))
	}

	// make the marshal.pb.go file
	pb.marshal(pf)

	// make the sort.pb.go file
	if err := pb.makeNewSortfile(pf); err != nil {
		badExit(err)
	}
}

func okExit(s string) {
	log.Info("autogenpb ok", s)
	os.Exit(0)
}

func badExit(err error) {
	log.Info("autogenpb error:", err)
	os.Exit(-1)
}
